Check 42.
Check 0.
Check 00.
Check 0x1ab.
Check 0X1ab.
Check 0x1Ab.
Check 0x1aB.
Check 0x1AB.
Fail Check 0x1ap1.  (* exponents not implemented (yet?) *)
Fail Check 0x1aP1.
Check 0x0.
Check 0x000.
Fail Check 0xg.
Fail Check 0xG.
Fail Check 00x1.
Fail Check 0x.
Open Scope hex_nat_scope.
Check 42.
