(* -*- coq-prog-args: ("-compat" "8.13") -*- *)
(** Check that the current-minus-two compatibility flag actually requires the relevant modules. *)
Import Coq.Compat.Coq815.
Import Coq.Compat.Coq814.
Import Coq.Compat.Coq813.
