Require Coq.ZArith.BinInt.
Definition foo0 := Eval vm_compute in Coq.ZArith.BinInt.Z.div_eucl.
Definition foo1 := Eval vm_compute in Coq.ZArith.BinInt.Z.div_eucl.
Definition foo2 := Eval vm_compute in Coq.ZArith.BinInt.Z.div_eucl.
