(*Doc, it hurts when I poke myself.*)

Notation "'" := 1. (* was:
Setting notation at level 0.
Toplevel input, characters 0-18:
> Notation "'" := 1.
> ^^^^^^^^^^^^^^^^^^
Anomaly: Uncaught exception Invalid_argument("index out of bounds"). Please report. *)
